/*
 * Copyright (C) 1998-2005 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "gui.cmdline.h"
#include "para.h"

#include <curses.h>


#define STANDARD_STATUS_BAR "para_gui " VERSION " (hit ? for help)"
#define MIN_TOP_LINES 9

static int signal_pipe[2];

const char *logo="       / \n"
"      /    _ __   __ _ _ __ __ _ ___| | __ _ ___| |__\n"
"  ####    | '_ \\ / _` | '__/ _` / __| |/ _` / __| '_ \\\n"
"  #  #    | |_) | (_| | | | (_| \\__ \\ | (_| \\__ \\ | | |\n"
"  ####    | .__/ \\__,_|_|  \\__,_|___/_|\\__,_|___/_| |_|\n"
" /        |_|\n"
"/";

static void finish(int sig);

static struct win_data {
	WINDOW *win;
	NCURSES_SIZE_T begx;
	NCURSES_SIZE_T begy;
	NCURSES_SIZE_T cols;
	NCURSES_SIZE_T lines;
} top, bot, sb, in, sep;

static int external_cmd_died, curses_active;
static pid_t external_cmd_pid;

static FILE *command_pipe, *stat_pipe;


static struct gengetopt_args_info args_info;

enum {GETCH_MODE, COMMAND_MODE, EXTERNAL_MODE};

#define LEFT 1
#define RIGHT 2
#define CENTER 3


#define COLOR_STATUSBAR 32
#define COLOR_COMMAND 33
#define COLOR_OUTPUT 34
#define COLOR_MSG 35
#define COLOR_ERRMSG 36
#define COLOR_WELCOME 37
#define COLOR_SEPARATOR 38

#define USECS 1000000
#define MSECS 1000

struct stat_item{
	char name[MAXLINE];
	char prefix[MAXLINE];
	char postfix[MAXLINE];
	unsigned y;
	unsigned x;
	unsigned len;
	int fg, bg;
	int align;
	char content[MAXLINE];
};

struct audio_format_info{
	char *name;
	char *rcmd, *wcmd;
	int bufsize;
	unsigned long penalty;
	struct timeval stime;
};

static struct audio_format_info afi[] = { /* audio format info */
#if 1
	{
		.name = "mp3",
		.rcmd = "pob-fec -q -b 2",
		.wcmd = "madplay -",
		.bufsize = 4096,
	},
#endif
#ifdef HAVE_OGGVORBIS
	{
		.name = "ogg",
		.rcmd = "ogg123 -b 32 -p 99 -d au http://localhost:8000 -f -",
		.wcmd = "aplay -",
		.bufsize = 4096,
	},
#endif
	{
		.name = NULL,
	}
};

struct stream_io {
	int format;
	FILE *read, *write;
	pid_t rpid, wpid;
	struct timeval rtime, wtime, stime;
	int buflen;
	char *buf;
	int rkilled, wkilled;
};

#define MAX_STREAM_SLOTS 10
static struct stream_io siol[MAX_STREAM_SLOTS];

enum {STATUS_BAR, STATUS, STATUS_FLAGS, NUM_PLAYED, STREAM, LENGTH_MIN, LENGTH_SEC,
	SCORE, MTIME, FILE_SIZE, FORMAT,
	AUDIO_INFO1, AUDIO_INFO2,AUDIO_INFO3,
	DBINFO1, DBINFO2, DBINFO3};

struct stat_item stat_items[] = {
	{
		.name = "status_bar",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_CYAN,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 7,
		.len = 100,
	},
	{
		.name = "status",
		.prefix = "status: ",
		.postfix = " ",
		.fg = COLOR_RED,
		.bg = COLOR_BLACK,
		.align = RIGHT,
		.content = "",
		.x = 0,
		.y = 17,
		.len = 35,
	},
	{
		.name = "status_flags",
		.prefix = "(",
		.postfix = ")",
		.fg = COLOR_RED,
		.bg = COLOR_BLACK,
		.align = LEFT,
		.content = "",
		.x = 35,
		.y = 17,
		.len = 10,
	},
	{
		.name = "num_played",
		.prefix = "#",
		.postfix = "",
		.fg = COLOR_RED,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 45,
		.y = 17,
		.len = 10,
	},
	{
		.name = "dbtool",
		.prefix = "dbtool: ",
		.postfix = "",
		.fg = COLOR_RED,
		.bg = COLOR_BLACK,
		.align = LEFT,
		.content = "",
		.x = 72,
		.y = 17,
		.len = 28,
	},
	{
		.name = "length_min",
		.prefix = "Length: ",
		.postfix = "",
		.fg = COLOR_MAGENTA,
		.bg = COLOR_BLACK,
		.align = RIGHT,
		.content = "",
		.x = 0,
		.y = 27,
		.len = 15,
	},
	{
		.name = "length_sec",
		.prefix = ":",
		.postfix = "",
		.fg = COLOR_MAGENTA,
		.bg = COLOR_BLACK,
		.align = LEFT,
		.content = "",
		.x = 15,
		.y = 27,
		.len = 8,
	},
	{
		.name = "uptime",
		.prefix = "Up: ",
		.postfix = "",
		.fg = COLOR_RED,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 55,
		.y = 17,
		.len = 17,
	},
	{
		.name = "mtime",
		.prefix = "Mtime: ",
		.postfix = "",
		.fg = COLOR_MAGENTA,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 23,
		.y = 27,
		.len = 40,
	},
	{
		.name = "file_size",
		.prefix = "Size: ",
		.postfix = " kb",
		.fg = COLOR_MAGENTA,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 63,
		.y = 27,
		.len = 20,
	},
	{
		.name = "format",
		.prefix = "format: ",
		.postfix = "",
		.fg = COLOR_MAGENTA,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 84,
		.y = 27,
		.len = 16,
	},
	{
		.name = "audio_file_info1",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_GREEN,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 43,
		.len = 100,
	},
	{
		.name = "audio_file_info2",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_GREEN,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 53,
		.len = 100,
	},
	{
		.name = "audio_file_info3",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_GREEN,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 63,
		.len = 100,
	},
	{
		.name = "dbinfo1",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_YELLOW,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 77,
		.len = 100,
	},
	{
		.name = "dbinfo2",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_YELLOW,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 87,
		.len = 100,
	},
	{
		.name = "dbinfo3",
		.prefix = "",
		.postfix = "",
		.fg = COLOR_YELLOW,
		.bg = COLOR_BLACK,
		.align = CENTER,
		.content = "",
		.x = 0,
		.y = 97,
		.len = 100,
	},
	{
		.name = "",
	}
};

int _argc;
char **_argv;

static void com_help(int);
static void com_reread_conf(int);
static void com_toggle_auto_decode(int);
static void com_enlarge_top_win(int);
static void com_shrink_top_win(int);
static void com_version(int);
static void com_quit(int);
static void com_refresh(int);
static void com_ll_incr(int);
static void com_ll_decr(int);

struct gui_command command_list[] = {
	{
		.key = "?",
		.name = "help",
		.description = "print help",
		.handler = com_help
	}, {
		.key = "+",
		.name = "enlarge_win",
		.description = "enlarge the top window",
		.handler = com_enlarge_top_win
	}, {
		.key = "-",
		.name = "shrink_win",
		.description = "shrink the top window",
		.handler = com_shrink_top_win
	}, {
		.key = "d",
		.name = "auto_decode",
		.description = "toggle auto decoding",
		.handler = com_toggle_auto_decode
	}, {
		.key = "r",
		.name = "reread_conf",
		.description = "reread configuration file",
		.handler = com_reread_conf
	}, {
		.key = "q",
		.name = "quit",
		.description = "kill decoder and exit",
		.handler = com_quit
	}, {
		.key = "^L",
		.name = "refresh",
		.description = "redraw the screen",
		.handler = com_refresh
	}, {
		.key = ">",
		.name = "ll_incr",
		.description = "increase loglevel (decreases verbosity)",
		.handler = com_ll_incr
	}, {
		.key = "<",
		.name = "ll_decr",
		.description = "decrease loglevel (increases verbosity)",
		.handler = com_ll_decr
	}, {
		.key = "V",
		.name = "version",
		.description = "show the para_gui version",
		.handler = com_version
	}, {
		.handler = NULL
	}
};

static int find_cmd_byname(char *name)
{
	int i;

	for (i = 0; command_list[i].handler; i++)
		if (!strcmp(command_list[i].name, name))
			return i;
	return -1;
}

/* taken from mutt */
static char *km_keyname (int c)
{
	static char buf[10];

	if (c < 256 && c > -128 && iscntrl((unsigned char) c)) {
		if (c < 0)
			c += 256;
		if (c < 128) {
			buf[0] = '^';
			buf[1] = (c + '@') & 0x7f;
			buf[2] = 0;
		} else
			snprintf(buf, sizeof(buf), "\\%d%d%d", c >> 6,
				(c >> 3) & 7, c & 7);
	} else if (c >= KEY_F0 && c < KEY_F(256))
		sprintf(buf, "<F%d>", c - KEY_F0);
	else if (isprint(c))
		snprintf(buf, sizeof(buf), "%c", (unsigned char) c);
	else
		snprintf(buf, sizeof(buf), "\\x%hx", (unsigned short) c);
	return buf;
}

static char *configfile_exists(void)
{
	static char *config_file;
	char *tmp;

	if (!args_info.config_file_given) {
		if (!config_file)
			config_file = malloc(_POSIX_PATH_MAX * sizeof(char));
		sprintf(config_file, "%s/.paraslash/gui.conf", getenv("HOME"));
		tmp = config_file;
	} else
		tmp = args_info.config_file_arg;
	return file_exists(tmp)? tmp: NULL;
}

/*
 * print num spaces to curses window
 */
static void add_spaces(WINDOW* win, unsigned int num)
{
	while (num > 0) {
		num--;
		waddstr(win, " ");
	}
}

/*
 * print aligned string to curses window. This function always prints
 * exactly len chars.
 */
static int align_str(WINDOW* win, char *string, unsigned int len,
		unsigned int align)
{
	int num;
	char *str;

	if (!win || !string)
		return -1;
	if (!(str = strdup(string)))
		return -1;
	num = len - strlen(str);	
	if (num < 0) {
		str[len] = '\0';
		num = 0;
	}
	if (align == LEFT) {
		waddstr(win, str);
		add_spaces(win, num);
	} else if (align == RIGHT) {
		add_spaces(win, num);
		waddstr(win, str);
	} else {
		add_spaces(win, num / 2);
		waddstr(win, str[0]? str: "");
		add_spaces(win, num - num / 2);
	}
	free(str);
	return 1;
}

static void print_in_bar(int color, char *fmt,...)
{
	va_list argp;
	static char buf[MAXLINE];

	if (!curses_active)
		return;
	wattron(in.win, COLOR_PAIR(color));
	va_start(argp, fmt);
	vsprintf(buf, fmt, argp);
	va_end(argp);
	wmove(in.win, 0, 0);
	align_str(in.win, buf? buf: " ", sb.cols, LEFT);
	wrefresh(in.win);
}

/*
 * update the status bar
 */
static void print_status_bar(char *line)
{
	static char slots[MAX_STREAM_SLOTS + 1];
	int i;
	char *tmp;

	if (!curses_active)
		return;
	for (i = 0; i < MAX_STREAM_SLOTS; i++)
		slots[i] = siol[i].format < 0? '_' : siol[i].format + '0';
	slots[i] = '\0';
	wmove(sb.win, 0, 0);
	tmp = make_message("[%s] " STANDARD_STATUS_BAR, slots);
//	align_str(sb.win, line? line: STANDARD_STATUS_BAR, sb.cols, CENTER);
	align_str(sb.win, tmp, sb.cols, CENTER);
	free(tmp);
	wrefresh(sb.win);
}

static void add_output_line(char *line)
{
	waddstr(bot.win, line);
	waddstr(bot.win, "\n");
	wrefresh(bot.win);
}

void clog(int ll, char *fmt,...)
{
	int color;
	va_list argp;
	static char buf[MAXLINE];

	if (ll < args_info.loglevel_arg)
		return;
	switch (ll) {
		case DEBUG:
		case VERBOSE:
		case INFO:
			color = COLOR_MSG;
			break;
		default:
			color = COLOR_ERRMSG;
	}
	va_start(argp, fmt);
	vsprintf(buf, fmt, argp);
	va_end(argp);
	if (curses_active) {
		wattron(bot.win, COLOR_PAIR(color));
		waddstr(bot.win, buf);
		wrefresh(bot.win);
	} else
		fprintf(stderr, "%s", buf);
}

static void setup_signal_handling(void)
{
	if (para_signal_init(signal_pipe, clog) < 0)
		exit(EXIT_FAILURE);
	para_install_sighandler(SIGINT);
	para_install_sighandler(SIGTERM);
	para_install_sighandler(SIGCHLD);
	para_install_sighandler(SIGWINCH);
	para_install_sighandler(SIGUSR1);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, SIG_IGN);
}

#if 0
static void kill_children(void)
{
	signal(SIGTERM, SIG_IGN);
	kill(0, SIGTERM);
	signal(SIGTERM, deadly_signal_handler);
}
#endif
static void do_exit(int ret)
{
	signal(SIGTERM, SIG_IGN);
	kill(0, SIGTERM);
	exit(ret);
}

static void shutdown_curses(void)
{
	if (!curses_active)
		return;
	def_prog_mode();
	curses_active = 0;
	endwin();
}

static void finish(int ret)
{
	shutdown_curses();
	do_exit(ret);
}

/*
 * exit curses and print given message
 */
static void msg_n_exit(int ret, char* fmt,...)
{
	va_list argp;
	FILE *outfd = ret? stderr: stdout;

	shutdown_curses();
	va_start(argp, fmt);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
	do_exit(ret);
}

/*
 * replace first '\n' by '\0' and return pointer to char after first ':'
 * in str.
 */
static char *prune_str(char *str, unsigned int len)
{
	char *p = str, *p1;
	unsigned int i = 0;

	while (*p != ':') {
		p++;
		if (*p == '\0')
			return NULL;
	}
	p++;
	p1 = p;
	while (*p1 != '\0' && i < len) {
		if (*p1 == '\n')
			break;
		p1++;
		i++;
	}
	*p1 = '\0';
	return p;
}

/*
 * print formated output to bot win
 */
static void outputf(int color, char* fmt,...)
{
	va_list argp;
	static char buf[MAXLINE];

	if (!curses_active)
		return;
	wattron(bot.win, COLOR_PAIR(color));
	va_start(argp, fmt);
	vsprintf(buf, fmt, argp);
	va_end(argp);
	waddstr(bot.win, buf);
	wrefresh(bot.win);
}

static void print_welcome(void)
{
	int ll = args_info.loglevel_arg;
	if (ll > NOTICE)
		return;
	outputf(COLOR_WELCOME, "Welcome to para_gui " VERSION
		" \"" CODENAME "\"\n");
	wclrtoeol(bot.win);
	if (ll < NOTICE && bot.lines > 23)
		outputf(COLOR_MSG, "%s\n", logo);
}

/*
 * int all windows
 */
static void init_wins(int top_lines)
{
	int i;

	top.lines = top_lines;
	top.cols = COLS;
	top.begy = 0;
	top.begx = 0;

	bot.lines = LINES - top.lines - 3;
	bot.cols = COLS;
	bot.begy = top.lines + 1;
	bot.begx = 0;

	sb.lines = 1;
	sb.cols = COLS;
	sb.begy = LINES - 2;
	sb.begx = 0;
	
	in.lines = 1;
	in.cols = COLS;
	in.begy = LINES - 1;
	in.begx = 0;

	sep.lines = 1;
	sep.cols = COLS;
	sep.begy = top.lines;
	sep.begx = 0;
	
	if (top.win) {
		mvwin(top.win, top.begy, top.begx);
		wresize(top.win, top.lines, top.cols);

		mvwin(sb.win, sb.begy, sb.begx);
		wresize(sb.win, sb.lines, sb.cols);

		mvwin(sep.win, sep.begy, sep.begx);
		wresize(sep.win, sep.lines, sep.cols);

		mvwin(bot.win, bot.begy, bot.begx);
		wresize(bot.win, bot.lines, bot.cols);

		mvwin(in.win, in.begy, in.begx);
		wresize(in.win, in.lines, in.cols);
	} else {
		sep.win = newwin(sep.lines, sep.cols, sep.begy, sep.begx);
		top.win = newwin(top.lines, top.cols, top.begy, top.begx);
		bot.win = newwin(bot.lines, bot.cols, bot.begy, bot.begx);
		sb.win = newwin(sb.lines, sb.cols, sb.begy, sb.begx);
		in.win = newwin(in.lines, in.cols, in.begy, in.begx);
		if (!top.win || !bot.win || !sb.win || !in.win || !sep.win)
			msg_n_exit(1, "Error: Cannot create curses windows\n");
		wclear(bot.win);
		wclear(top.win);
		wclear(sb.win);
		wclear(in.win);
		scrollok(bot.win, 1);
		wattron(sb.win, COLOR_PAIR(COLOR_STATUSBAR));
		wattron(sep.win, COLOR_PAIR(COLOR_SEPARATOR));
		nodelay(top.win, 1);
		nodelay(bot.win, 1);
		nodelay(sb.win, 1);
		nodelay(in.win, 0);

		keypad(top.win, 1);
		keypad(bot.win, 1);
		keypad(sb.win, 1);
		keypad(in.win, 1);
		print_status_bar(NULL);
	}
	
	for (i = 1; i <= COLS; i++)
		 waddstr(sep.win, "-");

	wclear(top.win);
	wnoutrefresh(top.win);
	wnoutrefresh(bot.win);
	//wnoutrefresh(sb.win);
	print_status_bar(NULL);
	wnoutrefresh(in.win);
	wnoutrefresh(sep.win);
	doupdate();
}

static int top_lines_ok(int top_lines)
{
	return (top_lines < MIN_TOP_LINES || LINES - top_lines < 5)? 0:1;
}

static void check_geometry(void)
{
	if (LINES < 13 || COLS < 60)
		msg_n_exit(EXIT_FAILURE, "Error: Terminal (%dx%d) too narrow"
			" (need at least 60x13)\n", COLS, LINES);
}

/*
 * Print item #i to curses window
 */
static void print_item(int i)
{
	struct stat_item *s = stat_items + i;
	char buf[MAXLINE];

	if (!curses_active)
		return;
	if (!s->content) {
		clog(ERROR, "s->content == NULL\n");
		return;
	}
	snprintf(buf, MAXLINE - 1, "%s%s%s", s->prefix, s->content, s->postfix);
	buf[MAXLINE] = '\0';
	wmove(top.win, s->y * top.lines / 100, s->x * COLS / 100);
	wrefresh(top.win);
	wattron(top.win, COLOR_PAIR(i + 1));
	align_str(top.win, buf, s->len * COLS / 100, s->align);
	wrefresh(top.win);
}

static void print_all_items(void)
{
	int i;

	if (!curses_active)
		return;
	for (i = 0; stat_items[i].name[0]; i++)
		print_item(i);
}

static void clear_all_items(void)
{
	int i;

	for (i = 0; stat_items[i].name[0]; i++)
		stat_items[i].content[0] = '\0';
}

static void init_colors(void)
{
	int i;
	if (!has_colors())
		msg_n_exit(EXIT_FAILURE, "Error: No color term\n");
	start_color();

	for (i = 0; stat_items[i].name[0]; i++)
		init_pair(i + 1, stat_items[i].fg, stat_items[i].bg);

	init_pair(COLOR_STATUSBAR, COLOR_BLACK, COLOR_GREEN);
	init_pair(COLOR_COMMAND, COLOR_YELLOW, COLOR_BLACK);
	init_pair(COLOR_OUTPUT, COLOR_CYAN, COLOR_BLACK);
	init_pair(COLOR_MSG, COLOR_WHITE, COLOR_BLACK);
	init_pair(COLOR_ERRMSG, COLOR_WHITE, COLOR_RED);
	init_pair(COLOR_WELCOME, COLOR_WHITE, COLOR_BLUE);
	init_pair(COLOR_SEPARATOR, COLOR_BLUE, COLOR_BLACK);
}

/*
 * (re-)initialize the curses library
 */
static void init_curses(void)
{
	//resize_term(LINES, COLS);
	refresh(); /* really needed */
	check_geometry();
	curs_set(0); /* really needed */
	noraw();
	nonl(); /* tell curses not to do NL->CR/NL on output */
	noecho(); /* don't echo input */
	cbreak(); /* take input chars one at a time, no wait for \n */
	//reset_prog_mode();
	init_colors();
	curses_active = 1;
	init_wins(top_lines_ok(top.lines)? top.lines : MIN_TOP_LINES);
	print_all_items();
	noecho(); /* don't echo input */
}

static long unsigned penalty_to_ms(int penalty)
{
	return 1 << (penalty + 8);
}

static void recalc_penalty(int format, long lifetime)
{
	if (lifetime > 30)
		afi[format].penalty = 0;
}

static void clear_slot(int slot)
{
	struct stream_io *s = &siol[slot];

	clog(INFO, "%s decoder in slot %d deactivated\n",
		afi[s->format].name, slot);
	s->format = -1;
	s->rkilled = 0;
	s->wkilled = 0;
	print_status_bar(NULL);
}

static void increase_penalty(int format)
{
	if (afi[format].penalty >= 8)
		return;
	afi[format].penalty++;
	clog(DEBUG, "increasing %s penalty to %lu (%lu ms)\n", 
		afi[format].name, afi[format].penalty,
		penalty_to_ms(afi[format].penalty));
}

static void check_sigchld(void)
{
	pid_t pid;
	int i;
reap_next_child:
//	clog(DEBUG, "reaping\n");
	pid = para_reap_child();
	if (pid <= 0)
		return;
	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		struct timeval now;
		struct stream_io *s = &siol[i];
		long lifetime;

		if (s->format < 0)
			continue;
		if (pid != s->rpid && pid != s->wpid)
			continue;
		gettimeofday(&now, NULL);
		if (pid == s->rpid) {
			lifetime = now.tv_sec - s->stime.tv_sec;
			clog(DEBUG, "%s stream reader in slot %d died "
				"after %d secs\n",
				afi[s->format].name,
				i,
				lifetime);
			if (lifetime < 2)
				increase_penalty(s->format);
			if (s->read) {
//				clog(DEBUG, "closing reading end of pipe\n");
				fclose(s->read);
//				clog(DEBUG, "pipe closed\n");
				s->read = NULL;
			}
			if (!s->write) {
				s->buflen = 0;
				clear_slot(i);
			} else
				recalc_penalty(s->format, lifetime);
			break;
		}
		lifetime = now.tv_sec - s->stime.tv_sec;
		clog(DEBUG, "%s stream writer in slot %d died "
			"after %d secs\n",
			afi[s->format].name, i, lifetime);
		if (lifetime < 2)
			increase_penalty(s->format);
		fclose(s->write);
		s->write = NULL;
		s->buflen = 0;
		if (!s->read)
			clear_slot(i);
		else
			recalc_penalty(s->format, lifetime);
		//kill_stream_reader(i);
//			usleep(300000);
		break;
	}
	if (pid == external_cmd_pid) {
		external_cmd_pid = 0;
		external_cmd_died = 1;
	}
	goto reap_next_child;
}

static int get_empty_slot(void)
{
	int i;
	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		if (siol[i].format < 0)
			return i;
	}
	return -1;
}

static int decoder_running(int format)
{
	int i;
	struct stream_io *s;

	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		s = &siol[i];
		if (s->format >= 0 && s->format == format && s->write && s->read)
			return 1;
	}
	return 0;
}

static void start_decoder(int format)
{
	struct audio_format_info *a = &afi[format];
	struct stream_io *s;
	int slot;

	slot = get_empty_slot();
	if (slot < 0)
		msg_n_exit(ERROR, "no more empty slots!\n");
	s = &siol[slot];
	clog(DEBUG, "starting %s decoder (r:%s, w:%s, slot %d)\n", a->name,
		a->rcmd, a->wcmd, slot);
	s->write = popen_write_cmdline_pid(&s->wpid, a->wcmd);
	/* we write to this fd in do_select, so we need non-blocking */
	fcntl(fileno(s->write), F_SETFL, O_NONBLOCK);
	gettimeofday(&s->wtime, NULL);
	clog(DEBUG, "%s stream_writer started, pid = %d\n", a->name, s->wpid);
	s->read = popen_read_cmdline_pid(&s->rpid, a->rcmd);
	clog(DEBUG, "%s stream reader started, pid = %d\n", a->name, s->rpid);
	s->buf = para_malloc(a->bufsize * sizeof(char)); /* leak? */
	s->buflen = 0;
	s->format = format;
	s->rkilled = 0;
	s->wkilled = 0;
	gettimeofday(&s->rtime, NULL);
	s->wtime = s->rtime;
	s->stime = s->rtime;
	a->stime = s->rtime;
	print_status_bar(NULL);
	clog(INFO, "%s decoder in slot %d activated\n", a->name, slot);
}

static int is_frozen(int format)
{
	struct audio_format_info *a = &afi[format];
	long difftime; /* milliseconds */
	struct timeval now;

	gettimeofday(&now, NULL);
	difftime = (now.tv_sec - a->stime.tv_sec) * 1000
		+ (now.tv_usec - a->stime.tv_usec) / 1000; 
	return penalty_to_ms(afi[format].penalty) > difftime? 1 : 0; 
}

static int start_current_decoder(void)
{
	int i;
	struct audio_format_info *a;

	if (!*stat_items[FORMAT].content)
		return -1;
	for (i = 0, a = afi; a->name; a++, i++)
		if (!strcmp(stat_items[FORMAT].content, a->name))
			break;
	if (!a->name)
		return -1;
	if (decoder_running(i))
		return 1;
	if (is_frozen(i))
		return -1;
	start_decoder(i);
	return 1;
}

static void kill_stream_writer(int slot)
{
	struct stream_io *s = &siol[slot];

	if (s->format < 0 || !s->write || s->wkilled)
		return;
	clog(DEBUG, "stopping %s stream writer (slot %d)\n",
		afi[s->format].name, slot);
	kill(s->wpid, SIGTERM);
	s->wkilled = 1;
	s->buflen = 0;
}

static void kill_stream_reader(int slot)
{
	struct stream_io *s = &siol[slot];

	if (s->format < 0 || !s->read || s->rkilled)
		return;
	clog(INFO, "stopping %s stream reader (slot %d)\n",
		afi[s->format].name, slot);
	s->rkilled = 1;
	kill(s->rpid, SIGTERM);
}

/*
 * Send SIGTERM to reader and writer if present
 */
static void kill_decoder(int slot)
{
//	check_sigchld();
	kill_stream_writer(slot);
	kill_stream_reader(slot);
//	usleep(100000);
}

static void kill_all_decoders(void)
{
	int i;

	for (i = 0; i < MAX_STREAM_SLOTS; i++)
		if (siol[i].format >= 0)
			kill_decoder(i);
}

/*
 * extract contents from line which is already known to start with
 * name of i-th command to stat_items' content.
 */
static void print_stat_line(char *line, int i)
{
	char *strptr;
	struct stat_item *s = stat_items + i;
	
	if (!line) {
		clog(ERROR, "%s: line == NULL\n", __func__);
		return;
	}
	strptr = prune_str(line, s->len - strlen(s->prefix)
		- strlen(s->postfix));
	if (!strptr)
		return;
	strcpy(s->content, strptr);
	if (curses_active)
		print_item(i);
}

/*
 * print status line if line starts with known command.
 */
static void check_stat_line(char *line)
{
	int i;
	size_t line_len = strlen(line);

	for (i = 0; stat_items[i].name[0]; i++) {
		struct stat_item *s = stat_items + i;
		size_t item_len = strlen(s->name);
		
		if (line_len > item_len && line[item_len] == ':' &&
 				!strncmp(line, s->name, item_len))
			print_stat_line(line, i);
	}
}

/*
 * This sucker modifies its first argument. *handler and *arg are
 * pointers to 0-terminated strings (inside line). Crap.
 */
static int split_key_map(char *line, char **handler, char **arg)
{
	if (!(*handler = strchr(line + 1, ':')))
		goto err_out;
	**handler = '\0';
	(*handler)++;
	if (!(*arg = strchr(*handler, ':')))
		goto err_out;
	**arg = '\0';
	(*arg)++;
	return 1;
err_out:
	return 0;
}

static int check_key_map_args(void)
{
	char *s;
	int i, ret = -1;
	char *tmp = NULL, *handler, *arg;

	for (i = 0; i < args_info.key_map_given; ++i) {
		s = args_info.key_map_arg[i];
		if (!(*s))
			goto err_out;
		if (tmp)
			free(tmp);
		tmp = strdup(s);
		if (!split_key_map(tmp, &handler, &arg))
			goto err_out;
		if (strlen(handler) != 1)
			goto err_out;
		if (*handler != 'x'
			&& *handler != 'd'
			&& *handler != 'i'
			&& *handler != 'p')
			goto err_out;
		if (*handler != 'i')
			continue;
		if (find_cmd_byname(arg) < 0)
			goto err_out;
	}
	ret = 0;
err_out:
	if (tmp)
		free(tmp);
	return ret;
}

/*
 * React to various signal-related events
 */
static void handle_signal(int sig)
{
	switch (sig) {
	case SIGTERM:
		msg_n_exit(EXIT_FAILURE,
			"only the good die young (caught SIGTERM))\n");
		return;
	case SIGWINCH:
		if (curses_active) {
			shutdown_curses();
			init_curses();
		}
		return;
	case SIGINT:
		clog(WARNING, "caught SIGINT, reset\n");
		/* Nothing to do. SIGINT killed our child, para_client stat.
		 * This get noticed by do_select which resets everything
		 */
		return;
	case SIGUSR1:
		clog(NOTICE, "got SIGUSR1, rereading configuration\n");
		com_reread_conf(0);
		return;
	case SIGCHLD:
		check_sigchld();
	}
}

static void open_stat_pipe_msg(void)
{
	stat_pipe = open_stat_pipe();
	clog(INFO, "connection to para_server established\n");
}

static void check_timeouts(void)
{
	struct timeval now;
	int slot, timeout = args_info.stream_timeout_arg;

	gettimeofday(&now, NULL);
	for (slot = 0; slot < MAX_STREAM_SLOTS; slot++) {
		struct stream_io *s = &siol[slot];
		if (s->format < 0)
			continue;
		/* check read time */
		if (s->read && !s->rkilled && 
			now.tv_sec > s->rtime.tv_sec + timeout) {
			clog(DEBUG, "%s input buffer (slot %d) not ready\n",
				afi[s->format].name, slot);
			kill_stream_reader(slot);
		}
		/* check write time */
		if (s->write && !s->wkilled && 
			now.tv_sec > s->wtime.tv_sec + timeout) {
			clog(DEBUG, "%s output buffer (slot %d) not ready\n",
				afi[s->format].name, slot);
			kill_stream_writer(slot);
		}
	}
}

static int set_stream_fds(fd_set *rfds, fd_set *wfds)
{
	int i, max_fileno = 0;
	if (!args_info.auto_decode_flag)
		return 0;
	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		struct stream_io *s = &siol[i];
		int file_no;

		if (s->format < 0)
			continue;
//		clog(DEBUG, "i = %d\n", i);
		if (s->buflen && s->write) {
			/*
                         * buf already contains data which was not yet
                         * witten
			 */
			file_no = fileno(s->write);
			FD_SET(file_no, wfds);
			max_fileno = MAX(file_no, max_fileno);
			continue;
		}
		if (s->read) {
			file_no = fileno(s->read);
			FD_SET(file_no, rfds);
			max_fileno = MAX(file_no, max_fileno);
		}
	}
	return max_fileno;
}

static int write_audio_data(int slot)
{
	struct stream_io *s = &siol[slot];
	int rv;

//	clog(DEBUG, "writing %d byte to %s pipe\n", s->buflen, afi[s->format].name);
	rv = write(fileno(s->write), s->buf, s->buflen);
	if (rv < 0) {
		clog(errno == EPIPE? INFO : ERROR, "write error: %s\n",
			strerror(errno));
		kill_decoder(slot);
		s->buflen = 0;	
	} else if (rv < s->buflen) {
		clog(DEBUG, "partial %s write (%i/%i) for slot %d\n",
			afi[s->format].name, rv, s->buflen, slot);
		s->buflen -= rv;
		memmove(s->buf, s->buf + rv, s->buflen);
	} else {
		gettimeofday(&s->wtime, NULL);
		s->buflen = 0;	
	}
	return rv;
}

static int read_audio_data(int slot)
{
	struct stream_io *s = &siol[slot];
	struct audio_format_info *a = &afi[s->format];
	int file_no = fileno(s->read);

//	clog(DEBUG, "reading at most %d byte from %s pipe (fileno %d)\n",
//			a->bufsize, a->name, file_no);
	s->buflen = read(file_no, s->buf, a->bufsize);
	if (s->buflen < 0) {
		clog(ERROR, "%s read error (%s)\n", a->name, strerror(errno));
		kill_decoder(slot);
		s->buflen = 0;
	} else if (s->buflen > 0)
		gettimeofday(&s->rtime, NULL);
	else
		kill_stream_reader(slot);
	return s->buflen;
}

static void do_stream_io(fd_set *rfds, fd_set *wfds)
{
	int i;
	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		struct stream_io *s = &siol[i];
		int file_no;
	
//		if (!s->write || !s->read || !args_info.auto_decode_flag)
//			continue;

		if (s->buflen) {
			if (!s->write)
				continue;
			file_no = fileno(s->write);
			if (FD_ISSET(file_no, wfds)) {
				write_audio_data(i);
				continue;
			}
		} else {
			if (!s->read)
				continue;
			file_no = fileno(s->read);
			if (FD_ISSET(file_no, rfds)) {
				read_audio_data(i);
				continue;
			}
		}
	}
}
static int stream_write_pending(void)
{
	int i;
	for (i = 0; i < MAX_STREAM_SLOTS; i++)
		if (siol[i].buflen > 0)
			return 1;
	return 0;
}
/*
 * This is the core function, dirty and mean. It has these modes:
 *
 * GETCH_MODE: check stdin and stat_pipe, return when key is
 * pressed
 *
 * COMMAND_MODE: check stat_pipe and command_pipe. Return when
 * a screen full of output has been read or when other end has closed command
 * pipe or when any key is pressed.
 *
 * EXTERNAL_MODE: Read/Write only stream pipe. Works with curses disabled.
 * Returns when external_cmd_pid == 0.
 */
static int do_select(int mode)
{
	fd_set rfds, wfds;
	/* stat pipe, command pipe, decoder */
	int sp_fileno, cp_fileno;
	int ret, lines = 0, total_lines = 0;
	int max_fileno, cp_numread = 1, sp_numread, playing;
	char status_buf[STRINGSIZE] = "";
	char command_buf[STRINGSIZE] = "";
	/* command/stat buf offset */
	int cbo = 0, sbo = 0;
	struct timeval tv;
repeat:
	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	max_fileno = 0;
	check_timeouts();
	playing = !strncmp(stat_items[STATUS].content, "playing", 7);
	if (!playing || !args_info.auto_decode_flag) {
		kill_all_decoders();
	} else {
		if (start_current_decoder() > 0)
			max_fileno = set_stream_fds(&rfds, &wfds);
		else
			playing = 0;
	}
	/* signal pipe */
	FD_SET(signal_pipe[0], &rfds);
	max_fileno = MAX(max_fileno, signal_pipe[0]);
	
	/* stat pipe */
	if (stat_pipe) {
		sp_fileno = fileno(stat_pipe);
		FD_SET(sp_fileno, &rfds);
		max_fileno = MAX(max_fileno, sp_fileno);
	}
	/* command pipe only for COMMAND_MODE */
	if (mode == COMMAND_MODE && lines < bot.lines - 1) {
		cp_fileno = fileno(command_pipe);
		FD_SET(cp_fileno, &rfds);
		max_fileno = MAX(max_fileno, cp_fileno);
	}
	tv.tv_sec = 0;
	tv.tv_usec = 200000;
	ret = select(max_fileno + 1, &rfds, &wfds, NULL, &tv);
	//clog(DEBUG, "select returned %d\n", ret);

	/* signals */
	if (FD_ISSET(signal_pipe[0], &rfds)) {
		int sig_nr = para_next_signal();
		if (sig_nr > 0)
			handle_signal(sig_nr);
	}
	if (ret <= 0)
		goto check_return; /* skip fd checks */
	if (playing)
		do_stream_io(&rfds, &wfds);
	/* read status pipe */
	if (stat_pipe && FD_ISSET(sp_fileno, &rfds)) {
		if ((sp_numread = read(sp_fileno, status_buf + sbo,
				STRINGSIZE - 1 - sbo)) <= 0) {
			//kill_all_decoders();
			clog(INFO, "closing server connection\n");
			pclose(stat_pipe);
			clear_all_items();
			print_all_items();
			stat_pipe = NULL;
			open_stat_pipe_msg();
			while (sleep(1) > 0) /* avoid busy loop */
				; /* try again*/
		}
		sbo = check_buf_for_items(status_buf, sp_numread + sbo,
			&check_stat_line, 0);
	}
	/* read command pipe if ready */
	if (mode == COMMAND_MODE && lines < bot.lines - 1 &&
			FD_ISSET(cp_fileno, &rfds)) {
		cp_numread = read(cp_fileno, command_buf + cbo,
			STRINGSIZE - 1 - cbo);
		if (cp_numread >= 0)
			cbo += cp_numread;
		else if (cp_numread < 0)
			clog(ERROR, "read error\n");
	}
check_return:
	if (playing && stream_write_pending())
		goto repeat;
	switch (mode) {
	case COMMAND_MODE:
		if (cp_numread <= 0 && cbo == 0) /* command complete */
			return 0;
		if (cbo && lines < bot.lines - 1) {
			int num = check_buf_for_items(command_buf, cbo,
				NULL, 0);
			if (num > 0) {
				num = MIN(num, bot.lines - 1 - lines);
				cbo = check_buf_for_items(command_buf, cbo,
					&add_output_line, num);
				lines += num;
			}
			if (lines >= bot.lines - 1)
				print_in_bar(COLOR_MSG,
					"Line %i-%i. Hit q to quit, any other "
					"key for next page", total_lines + 1,
					total_lines + lines);
		}
		ret = wgetch(top.win);
		if (ret != ERR && ret != KEY_RESIZE) {
			total_lines += lines;
			lines = 0;
		}
		if (ret == 'q') {
			cbo = 0;
			return -1;
		}
		break;
	case GETCH_MODE:
		ret = wgetch(top.win);
		if (ret != ERR && ret != KEY_RESIZE)
			return ret;
		break;
	case EXTERNAL_MODE:
		if (external_cmd_died) {
			external_cmd_died = 0;
			return 0;
		}
	}
	goto repeat;
}

/*
 * read from command pipe and print data to bot window
 */
static int send_output(pid_t pid)
{
	if (!command_pipe)
		return 0;
	if (do_select(COMMAND_MODE) >= 0)
		clog(INFO, "command complete\n");
	else
		clog(NOTICE, "command aborted\n");
	fclose(command_pipe);
	command_pipe = NULL;
	print_in_bar(COLOR_MSG, " ");
	return 1;
}

/*
 * exec client command and print output to status bar.
 */
static void print_executing_msg(char *cmd)
{
	outputf(COLOR_COMMAND, "%s\n", cmd);
	print_in_bar(COLOR_MSG, "executing client command, hit q to abort\n",
		cmd);
	wattron(bot.win, COLOR_PAIR(COLOR_OUTPUT));
}

#if 0
static int client_cmd(char *cmd,...)
{
	pid_t pid;
	print_executing_msg(cmd);
	command_pipe = popen_read_client_pid(&pid, &cmd);
	return send_output(pid);
}
#endif

static int client_cmd_cmdline(char *cmdline)
{
	pid_t pid;

	print_executing_msg(cmdline);
	command_pipe = popen_read_client_cmdline_pid(&pid, cmdline);
	return send_output(pid);
}

/*
 * exec command and print output to bot win
 */
static int display_cmd(char *cmd)
{
	pid_t pid;

	print_in_bar(COLOR_MSG, "executing display command, hit q to abort");
	outputf(COLOR_COMMAND, "%s\n", cmd);
	wattron(bot.win, COLOR_PAIR(COLOR_OUTPUT));

	command_pipe = popen_read_cmdline_pid(&pid, cmd);
	if (command_pipe)
		return send_output(pid);
	return -1;
}

/*
 * shutdown curses and stat pipe before executing external commands
 */
static int external_cmd(char *cmd)
{
	if (external_cmd_pid)
		return -1;
	shutdown_curses();
	exec_cmdline_pid_bg(&external_cmd_pid, cmd);
	do_select(EXTERNAL_MODE);
	init_curses();
	return 0;
}

static void com_ll_decr(int key)
{
	if (args_info.loglevel_arg <= DEBUG) {
		print_in_bar(COLOR_ERRMSG,
			"loglevel already at maximal verbosity\n");
		return;
	}
	args_info.loglevel_arg--;
	print_in_bar(COLOR_MSG, "loglevel set to %d\n", args_info.loglevel_arg);
}

static void com_ll_incr(int key)
{
	if (args_info.loglevel_arg >= EMERG) {
		print_in_bar(COLOR_ERRMSG,
			"loglevel already at miminal verbosity\n");
		return;
	}
	args_info.loglevel_arg++;
	print_in_bar(COLOR_MSG, "loglevel set to %d\n", args_info.loglevel_arg);
}

static int init_stream_io(void)
{
	int i, j, ret = -1;
	char *p, *q = NULL;

	i = args_info.stream_write_cmd_given - 1;
	while (i >= 0) {
		q = strdup(args_info.stream_write_cmd_arg[i]);
		p = strchr(q, ':');
		if (!p) {
			fprintf(stderr, "%s: invalid stream_write_cmd\n",
				__func__);
			goto out;
		}
		*p = '\0';
		p++;
		for (j = 0; afi[j].name; j++) {
			if (strcmp(q, afi[j].name))
				continue;
			afi[j].wcmd = args_info.stream_write_cmd_arg[i]
				+ (p - q);
			break;
		}
		free(q);
		q = NULL;
		i--;
	}
	//fprintf(stderr, "%s: invalid stream_read_cmd\n", __func__);
	i = args_info.stream_read_cmd_given - 1;
	//fprintf(stderr, "%s: %d\n", __func__, args_info.stream_read_cmd_given);
	while (i >= 0) {
		q = strdup(args_info.stream_read_cmd_arg[i]);
		p = strchr(q, ':');
		if (!p) {
			fprintf(stderr, "%s: invalid stream_read_cmd\n",
				__func__);
			goto out;
		}
		*p = '\0';
		p++;
		for (j = 0; afi[j].name; j++) {
			if (strcmp(q, afi[j].name))
				continue;
			afi[j].rcmd = args_info.stream_read_cmd_arg[i]
				+ (p - q);
			break;
		}
		free(q);
		q = NULL;
		i--;
	}
	ret = 1;
out:
	if (q)
		free(q);
	return ret;
}

/*
 * reread configuration, terminate on errors
 */
static void com_reread_conf(int key)
{
	char *cf =configfile_exists();

	if (!cf) {
		clog(WARNING, "there is no configuration to read\n");
		return;
	}
	clog(INFO, "rereading command line options and config file\n");
	cmdline_parser(_argc, _argv, &args_info);
	cmdline_parser_configfile(cf, &args_info, 1, 1, 0);
	clog(NOTICE, "configuration read\n");
	if (check_key_map_args() < 0)
		finish(EXIT_FAILURE);
	if (init_stream_io() < 0)
		finish(EXIT_FAILURE);
}

static void com_help(int key)
{
	int i, j = 0;
	char keymaps[MAXLINE], *p = keymaps;

	for (i = 0; i < args_info.key_map_given; ++i) {
		char *s = args_info.key_map_arg[i], *handler, *arg,
			*desc = "", tmp[MAXLINE];
		
		strcpy(tmp, s);
		if (!split_key_map(tmp, &handler, &arg))
			return;
		switch(*handler) {
			case 'i':
				handler = "internal";
				desc = command_list[find_cmd_byname(arg)].description;
				break;
			case 'x': handler = "external"; break;
			case 'd': handler = "display "; break;
			case 'p': handler = "para    "; break;
		}
		outputf(COLOR_MSG, "%s\t%s\t%s%s\t%s\n", tmp, handler, arg,
			strlen(arg) < 8? "\t" : "",
			desc);
		if (i > 0 && !(i % (bot.lines - 1))) {
			print_in_bar(COLOR_MSG, "Hit q to abort, any other "
				"key for next page");
			if (do_select(GETCH_MODE) == 'q') {
				print_in_bar(COLOR_MSG, "Help aborted");
				return;
			}
		}
	}
	*p = '\0';
	j = i;
	for (i = 0; command_list[i].handler; i++) {
		struct gui_command gc = command_list[i];
		
		outputf(COLOR_MSG, "%s\tinternal\t%s\t%s%s\n", gc.key, gc.name,
			strlen(gc.name) < 8? "\t" : "",
			gc.description);
		j++;
		if (!(j % (bot.lines - 1)))
			if (do_select(GETCH_MODE) < 0)
				return;
	}
	print_in_bar(COLOR_MSG, "try :help or :help command for more info");
	return;
}


static void com_toggle_auto_decode(int key)
{
	int ad = args_info.auto_decode_flag;

	ad = ad? 0 : 1;
		
	if (ad)
		clog(NOTICE, "activating auto decoding\n");
	else
		clog(NOTICE, "deactivating auto decoding\n");
	args_info.auto_decode_flag = ad;
}

static void com_shrink_top_win(int key)
{
	if (!top_lines_ok(top.lines - 1)) {
		clog(WARNING, "can not decrease top window\n");
		return;
	}
	init_wins(top.lines - 1);
	wclear(top.win);
	print_all_items();
	print_in_bar(COLOR_MSG, "decreased top window");
}

static void com_enlarge_top_win(int key)
{
	if (!top_lines_ok(top.lines + 1)) {
		clog(WARNING, "can not increase top window");
		return;
	}
	init_wins(top.lines + 1);
	wclear(top.win);
	print_all_items();
	print_in_bar(COLOR_MSG, "increased top window");
}

static void com_version(int key)
{
	print_in_bar(COLOR_MSG, "para_gui " VERSION " \"" CODENAME "\"");
}

static void com_quit(int key)
{
	finish(0);
}

static void com_refresh(int key)
{
	shutdown_curses();
	init_curses();
}

static void handle_command(int c)
{
	int i;

	/* first check user's key bindings */
	for (i = 0; i < args_info.key_map_given; ++i) {
		char tmp[MAXLINE], *handler, *arg;

		strcpy(tmp, args_info.key_map_arg[i]);
		if (!split_key_map(tmp, &handler, &arg))
			return;
		if (!strcmp(tmp, km_keyname(c))) {
			if (*handler == 'd') {
				display_cmd(arg);
				return;
			}
			if (*handler == 'x') {
				external_cmd(arg);
				return;
			}
			if (*handler == 'p') {
				client_cmd_cmdline(arg);
				return;
			}
			if (*handler == 'i') {
				int num = find_cmd_byname(arg);
				if (num >= 0)
					command_list[num].handler(c);
				return;
			}
		}
	}
	/* not found, check internal key bindings */
	for (i = 0; command_list[i].handler; i++) {
		if (!strcmp(km_keyname(c), command_list[i].key)) {
			command_list[i].handler(c);
			return;
		}
	}
	
	//outputf(COLOR_ERRMSG, "Unknown key: %i\n", c);
	print_in_bar(COLOR_ERRMSG, "key is not bound, press ? for help");
}


static void init_slots(void)
{
	int i;
	
	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		siol[i].read = NULL;
		siol[i].write = NULL;
		siol[i].format = -1;
		siol[i].rkilled = 0;
		siol[i].wkilled = 0;
	}
}

int main(int argc, char *argv[])
{
	int ret;
	char *cf;

	_argc = argc;
	_argv = argv;

	if (cmdline_parser(argc, argv, &args_info)) {
		fprintf(stderr, "parse error while reading command line\n");
		exit(EXIT_FAILURE);
	}
	if (check_key_map_args() < 0) {
		fprintf(stderr, "invalid key map\n");
		exit(EXIT_FAILURE);
	}
	cf = configfile_exists();
	if (!cf && args_info.config_file_given) {
		fprintf(stderr, "can not read config file %s\n",
			args_info.config_file_arg);
		exit(EXIT_FAILURE);
	}
	if (cf)
		cmdline_parser_configfile(cf, &args_info, 0, 0, 0);
	if (check_key_map_args() < 0) {
		fprintf(stderr, "invalid key map in config file\n");
		exit(EXIT_FAILURE);
	}
	if (init_stream_io() < 0) {
		fprintf(stderr, "invalid stream_read or stream_write command\n");
		exit(EXIT_FAILURE);
	}
	top.lines = args_info.top_lines_arg;
	setup_signal_handling();
	init_slots();
	initscr(); /* needed only once */
	init_curses();
	print_welcome();
	open_stat_pipe_msg();
	for (;;) {
		print_status_bar(NULL);
		ret = do_select(GETCH_MODE);
		if (!ret)
			continue;
		print_in_bar(COLOR_MSG, " ");
		handle_command(ret);
	}
}
