/*
 * Copyright (C) 2004-2005 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include <ogg/ogg.h>
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>

#include "server.h"

#define CHUNK_SIZE 8192
#define CHUNK_TIME 0.5

static OggVorbis_File *oggvorbis_file;
static FILE *infile;
static int chunks_sent, num_chunks, chunk_offset;
static struct timeval stream_start;
static ogg_int64_t *chunk_table, max_chunk_len;
struct audio_format *af;

/*
 * Alloc and fill array table of byte offsets. chunk_table[i] is the
 * offset in the current infile at which the sample containing time i *
 * CHUNK_TIME begins.
 */
static void ogg_compute_chunk_table(double time_total, ogg_int64_t raw_total)
{
	int i, ret = 0, num;
	ogg_int64_t pos, min, old_pos = 0;

	if (chunk_table)
		free(chunk_table);
	num = time_total / CHUNK_TIME + 2;
	clog(DEBUG, "%s: total time: %g allocating %d chunk pointers\n", 
		__func__, time_total, num);
	chunk_table = para_malloc(num * sizeof(ogg_int64_t));
	chunk_table[0] = 0;
	max_chunk_len = 0;
	rewind(infile);
	for (i = 1; ret == 0; i++) {
		ogg_int64_t diff;
		ret = ov_time_seek(oggvorbis_file, i * CHUNK_TIME);
		if (ret)
			break;
		pos = ov_raw_tell(oggvorbis_file);
		diff = pos - old_pos;
		max_chunk_len = MAX(max_chunk_len, diff);
		min = (i == 1)? diff : MIN(min, diff);
		old_pos = pos;
		chunk_table[i] = pos;
		if (i > 5 && i % 100 && i < num - 3)
			continue;
		clog(DEBUG, "%s: #%d: %g secs = %lli bytes\n", __func__, i, 
			i * CHUNK_TIME, pos);
	}
	chunk_table[i] = pos;
	num_chunks = i - 1;
	clog(DEBUG, "%s: %d chunks, max chunk: %lli, min chunk: %lli\n", 
		__func__, num_chunks, max_chunk_len, min);
	rewind(infile);
}

/*
 * Init oggvorbis file and write some tech data to given pointers.
 */
static int ogg_get_file_info(FILE *file, char *info_str, long unsigned *frames,
	int *seconds)
{
	int ret;
	double time_total;
	vorbis_info *vi;
	ogg_int64_t raw_total;
	
	infile = file;
	if (!file) {
		clog(CRIT, "%s: invalid file pointer\n", __func__);
		return -1;
	}
	rewind(file);
	oggvorbis_file = para_malloc(sizeof(OggVorbis_File));
	ret = ov_open(file, oggvorbis_file, NULL, 0);
	if (ret < 0) {
		clog(ERROR, "%s: ogg format not recognized\n", __func__);
		free(oggvorbis_file);
		return ret;
	}
	vi = ov_info(oggvorbis_file, 0);
	if (!vi)
		return -1;
	time_total = ov_time_total(oggvorbis_file, -1);
	raw_total = ov_raw_total(oggvorbis_file, -1);
	*seconds = time_total;
	*frames = ov_time_total(oggvorbis_file, -1) * 1000;
	*seconds = *frames / 1000;
	sprintf(info_str, "audio_file_info1:%ldkHz, %d channels, %ldkbps, "
		"%lu samples\n"
		"audio_file_info2: \n"
		"audio_file_info3: \n",
		vi->rate / 1000, vi->channels, ov_bitrate(oggvorbis_file, 0) / 1000,
		*frames);
	chunks_sent = 0;
	chunk_offset = 0;
	ogg_compute_chunk_table(time_total, raw_total);
	return 1;
}

/*
 * Called exactly once for each file.
 */
static void ogg_end_of_input(void)
{
	if (oggvorbis_file) {
		clog(INFO, "%s: closing vorbis file\n", __func__);
		ov_clear(oggvorbis_file);
		oggvorbis_file = NULL;
	}
	if (af->fifo >= 0) {
		clog(INFO, "%s: closing fifo %d\n", __func__, af->fifo);
		close(af->fifo); /* you are going to suck void, buddy */
		af->fifo = -1;
	}
	chunks_sent = 0;
}

/*
 * Simple stream reposition routine
 */
static int ogg_repos(long unsigned request, struct timeval *stream_start)
{
	int ret, offset;
	long seek;

	offset = request / CHUNK_TIME / 1000;
	seek = chunk_table[offset];
	ret = fseek(infile, seek, SEEK_SET);
	clog(DEBUG, "%s: seek to %d returned %d. Offset:%d\n", __func__, seek, ret,
		ftell(infile));
	if (ret)
		goto out;
	clog(DEBUG, "%s: using offset: %d/%d\n", __func__, offset, num_chunks);
	chunks_sent = 0;
	chunk_offset = offset;
out:
	return ret;
}

static int num_void_chunks(void)
{
	int vc, c = chunks_sent + chunk_offset;

	for (vc = 0; c + vc <= num_chunks; vc++)
		if (chunk_table[c + vc + 1] != chunk_table[c + vc])
			break;
	return c + vc > num_chunks? -1 : vc;
}

/*
 * Main working function.
 */
static int ogg_send_chunk(char *info_str, long unsigned *chunk_num,
	int afs_status)
{
	static int bytes, buflen;
	int ret, err, void_chunks;
	static struct timeval a;
	static char *buf;
	ogg_int64_t chunk_size;

	if (afs_status & AFS_NEXT) {
		bytes = 0;
		return 0;
	}
	if (af->fifo < 0) {
		bytes = 0;
		af->fifo = afs_open_fifo(af->fifoname);
		if (af->fifo < 0)
			return -1;
	}
	if (bytes > 0) {
		clog(INFO, "%s: completing short write (%d bytes not "
			"yet written)\n", __func__);
		goto write;
	}
	if (!buf || buflen < max_chunk_len) {
		clog(INFO, "%s: increasing ogg buffer size (%d -> %d)\n", 
			__func__, buflen, max_chunk_len);
		buf = para_realloc(buf, max_chunk_len);
		buflen = max_chunk_len;
	}
	if (afs_status & AFS_REPOS)
		ogg_repos(*chunk_num, &stream_start);
	chunk_size = 0;
	void_chunks = num_void_chunks();
	if (void_chunks < 0)
		return 0;
	chunk_size = chunk_table[chunks_sent + void_chunks + 1 + chunk_offset] 
		- chunk_table[chunks_sent + void_chunks + chunk_offset];
	if (!chunk_size)
		chunk_size = buflen;
	clog(DEBUG, "%s: reading chunk #%d/%d (%lli bytes)\n", __func__, 
		chunks_sent + 1 + chunk_offset, num_chunks, chunk_size);
        bytes = para_fread(buf, chunk_size, 1, infile);
	if (bytes <= 0) {
		clog(DEBUG, "%s: read returned %d\n", __func__, bytes);
		ret = bytes;
		bytes = 0;
		return ret;
	}
	if (bytes && bytes < chunk_size) {
		clog(CRIT, "%s: oops, short read (%d/%d)\n", __func__, 
			bytes, chunk_size);
		bytes = 0;
		return -1;
	}
write:
//	clog(DEBUG, "%s: writing %d bytes to fifo %d (%d:%d:%d)\n", __func__, bytes, 
//		fifo, *buf, *(buf + 1), *(buf + 2));
	if (!chunks_sent) {
		gettimeofday(&a, NULL);
		gettimeofday(&stream_start, NULL);
	}
	ret = write(af->fifo, buf, bytes);
	err = errno;
	if (ret <= 0) {
		clog(INFO, "%s: write failed (%s), %d, %d\n", 
			__func__, 
			strerror(err),
			err, 
			EAGAIN
);
		bytes = 0;
		return ret;
	}
	if (ret < bytes) {
		clog(WARNING, "%s: short write\n", __func__);
		memcpy(buf, buf + ret, bytes - ret);
		bytes -= ret;
		return 1;
	}
	*chunk_num = (chunks_sent + chunk_offset) * CHUNK_TIME * 1000;
	chunks_sent += void_chunks + 1;
	bytes -= ret;
	return 1;
}

static int ogg_mainloop(int current_af_is_ogg, unsigned int afs_status, 
	struct timeval **tvptr)
{
	unsigned next = afs_status & AFS_NEXT;
	unsigned playing = afs_status & AFS_PLAYING;
	static struct timeval tv;
	struct timeval tmp;

	*tvptr = NULL;
	if (next && af->sw_pid && af->fifo > 0)
		ogg_end_of_input();
	if (playing && !af->sw_pid && current_af_is_ogg && oggvorbis_file) {
		if (exec_cmdline_pid_bg(&af->sw_pid, af->open_cmd) < 0)
			return -1;
		if (af->fifo < 0)
			af->fifo = afs_open_fifo(af->fifoname);
	}
	if (!afs_status) {
		/* paused */
		if (chunks_sent) {
			chunk_offset = chunks_sent + chunk_offset;
			chunks_sent = 0;
		}
	}
	if (!playing || !oggvorbis_file || af->fifo < 0 || !af->sw_pid)
		return -1;
	if (!chunks_sent)
		*tvptr = NULL;
	else {
		d2tv(chunks_sent * CHUNK_TIME, &tmp); 
		tv_add(&tmp, &stream_start, &tv);
		*tvptr = &tv;
	}
	return af->fifo;
}

int ogg_init(void *p)
{
	char *open_cmd;
	af = p;
	af->get_file_info = ogg_get_file_info,
	af->send_chunk = ogg_send_chunk;
	af->mainloop = ogg_mainloop;
	af->fifo = -1;
	if (!af->open_cmd) {
		open_cmd = make_message("para_ovsend -r %s", af->fifoname);
		af->open_cmd = open_cmd;
	}
	gettimeofday(&stream_start, NULL);
	clog(INFO, "%s: opening stream writer %s\n", __func__, af->open_cmd);
	return 1;
}
